from typing import Optional

from famapy.metamodels.configuration_metamodel.models.configuration import Configuration

from famapy.metamodels.bdd_metamodel.models import BDDModel
from famapy.metamodels.bdd_metamodel.operations.interfaces import FeatureInclusionProbability
from famapy.metamodels.bdd_metamodel.operations import BDDProducts


class BDDFeatureInclusionProbabilityBF(FeatureInclusionProbability):
    """The Feature Inclusion Probability (FIP) operation determines the probability
    for a variable to be included in a valid solution.

    This is a brute-force implementation that enumerates all solutions
    for calculating the probabilities.

    Ref.: [Heradio et al. 2019. Supporting the Statistical Analysis of Variability Models. SPLC.
    (https://doi.org/10.1109/ICSE.2019.00091)]
    """

    def __init__(self, partial_configuration: Configuration = None) -> None:
        self.bdd_model = None
        self.result: dict[str, float] = {}
        self.partial_configuration = partial_configuration

    def execute(self, model: BDDModel) -> 'BDDFeatureInclusionProbabilityBF':
        self.bdd_model = model
        self.result = feature_inclusion_probability(self.bdd_model, self.partial_configuration)
        return self

    def get_result(self) -> dict[str, float]:
        return self.result

    def feature_inclusion_probability(self) -> dict[str, float]:
        return feature_inclusion_probability(self.bdd_model, self.partial_configuration)


def feature_inclusion_probability(bdd_model: BDDModel,
                                  config: Optional[Configuration] = None) -> dict[str, float]:
    products = BDDProducts(config).execute(bdd_model).get_result()
    n_products = len(products)
    if n_products == 0:
        return {feature: 0.0 for feature in bdd_model.variables}

    prob = {}
    for feature in bdd_model.variables:
        prob[feature] = sum(feature in p.elements for p in products) / n_products
    return prob



from typing import Optional

from famapy.metamodels.configuration_metamodel.models.configuration import Configuration

from famapy.core.operations import Products

from famapy.metamodels.bdd_metamodel.models.bdd_model import BDDModel


class BDDProducts(Products):
    """It computes all the solutions of a BDD model.

    It also supports the computation of all solutions from a partial configuration.
    """

    def __init__(self, partial_configuration: Configuration = None) -> None:
        self.result: list[Configuration] = []
        self.bdd_model = None
        self.partial_configuration = partial_configuration

    def execute(self, model: BDDModel) -> 'BDDProducts':
        self.bdd_model = model
        self.result = products(self.bdd_model, self.partial_configuration)
        return self

    def get_result(self) -> list[Configuration]:
        return self.result

    def get_products(self) -> list[Configuration]:
        return products(self.bdd_model, self.partial_configuration)


def products(bdd_model: BDDModel,
             partial_config: Optional[Configuration] = None) -> list[Configuration]:
    if partial_config is None:
        u_func = bdd_model.root
        care_vars = bdd_model.variables
        elements = {}
    else:
        values = dict(partial_config.elements.items())
        u_func = bdd_model.bdd.let(values, bdd_model.root)
        care_vars = set(bdd_model.variables) - values.keys()
        elements = partial_config.elements

    configs = []
    for assignment in bdd_model.bdd.pick_iter(u_func, care_vars=care_vars):
        features = {f: True for f in assignment.keys() if assignment[f]}
        features = features | elements
        configs.append(Configuration(features))
    return configs



from typing import Optional
from enum import Enum

from dd.autoref import Function

from famapy.core.transformations import ModelToText

from famapy.metamodels.bdd_metamodel.models.bdd_model import BDDModel


class BDDDumpFormat(Enum):
    """Possible output format for representing a BDD."""
    DDDMP_V3 = 'dddmp'
    DDDMP_V2 = 'dddmp2'
    PDF = 'pdf'
    PNG = 'png'
    SVG = 'svg'


class BDDWriter(ModelToText):
    """Create the dump file representing the argument BDD.

    The format can be:
        - dddmp v3: `'.dddmp'` (default)
        - dddmp v2: `'.dddmp2'`
        - PDF: `'.pdf'`
        - PNG: `'.png'`
        - SVG: `'.svg'`
    """

    @staticmethod
    def get_destination_extension() -> str:
        return BDDDumpFormat.DDDMP_V3.value

    def __init__(self, path: str, source_model: BDDModel, roots: Optional[list[Function]] = None, 
                 output_format: BDDDumpFormat = BDDDumpFormat.DDDMP_V3) -> None:
        self._path = path
        self._source_model = source_model
        self._output_format = output_format
        self._roots = roots

    def set_format(self, output_format: BDDDumpFormat) -> None:
        self._output_format = output_format

    def set_roots(self, roots: list[Function]) -> None:
        self._roots = roots

    def transform(self) -> None:
        self._source_model.bdd.dump(filename=self._path, roots=self._roots)
        if self._output_format == BDDDumpFormat.DDDMP_V3:
            # Convert to dddmp format version 3.0 (adding the '.varnames' field)
            dddmp_v2_to_v3(self._path)


def dddmp_v2_to_v3(filepath: str) -> None:
    """Convert the file with the BDD dump in format dddmp version 2 to version 3.

    The difference between versions 2.0 and 3.0 is the addition of the '.varnames' field.
    """

    with open(filepath, 'r', encoding='utf-8') as file:
        lines = file.readlines()
        # Change version from 2.0 to 3.0
        i, line = next((i, l) for i, l in enumerate(lines) if '.ver DDDMP-2.0' in l)
        lines[i] = line.replace('2.0', '3.0')

        # Add '.varnames' field
        i, line = next((i, l) for i, l in enumerate(lines) if '.orderedvarnames' in l)
        lines.insert(i - 1, line.replace('.orderedvarnames', '.varnames'))

    with open(filepath, 'w', encoding='utf-8') as file:
        file.writelines(lines)



import random
from typing import Optional

from famapy.metamodels.configuration_metamodel.models.configuration import Configuration
from famapy.core.operations import Sampling

from famapy.metamodels.bdd_metamodel.models import BDDModel
from famapy.metamodels.bdd_metamodel.operations import BDDProductsNumber


class BDDSampling(Sampling):
    """Uniform Random Sampling (URS) using a Binary Decision Diagram (BDD).

    This is an adaptation of
    [Heradio et al. 2021.
    Uniform and Scalable Sampling of Highly Configurable Systems.
    Empirical Software Engineering]
    which relies on counting-based sampling inspired in the original Knuth algorithm.

    This implementation supports samples with and without replacement,
    as well as samples from a given partial configuration.
    """

    def __init__(self, size: int, with_replacement: bool = False,
                 partial_configuration: Optional[Configuration] = None) -> None:
        self.result: list[Configuration] = []
        self.bdd_model = None
        self.size = size
        self.with_replacement = with_replacement
        self.partial_configuration = partial_configuration

    def execute(self, model: BDDModel) -> 'BDDSampling':
        self.bdd_model = model
        self.result = sample(self.bdd_model, self.size, self.with_replacement,
                             self.partial_configuration)
        return self

    def get_result(self) -> list[Configuration]:
        return self.result

    def sample(self, size: int, with_replacement: bool = False,
               partial_configuration: Optional[Configuration] = None) -> list[Configuration]:
        return sample(self.bdd_model, size, with_replacement, partial_configuration)


def sample(bdd_model: BDDModel, size: int, with_replacement: bool = False,
           partial_configuration: Optional[Configuration] = None) -> list[Configuration]:
    nof_configs = BDDProductsNumber(partial_configuration).execute(bdd_model).get_result()
    if size < 0 or (size > nof_configs and not with_replacement):
        raise ValueError('Sample larger than population or is negative.')

    configurations = []
    for _ in range(size):
        config = random_configuration(bdd_model, partial_configuration)
        configurations.append(config)

    if not with_replacement:
        set_configurations = set(configurations)
        while len(set_configurations) < size:
            config = random_configuration(bdd_model, partial_configuration)
            set_configurations.add(config)

    return list(set_configurations)


def random_configuration(bdd_model: BDDModel,
                         p_config: Optional[Configuration] = None) -> Configuration:
    # Initialize the configurations and values for BDD nodes with already known features
    values = {} if p_config is None else dict(p_config.elements.items())

    # Set the BDD nodes with the already known features values
    u_func = bdd_model.bdd.let(values, bdd_model.root)

    care_vars = set(bdd_model.variables) - values.keys()
    n_vars = len(care_vars)
    for feature in care_vars:
        # Number of configurations with the feature selected
        v_sel = bdd_model.bdd.let({feature: True}, u_func)
        nof_configs_var_selected = bdd_model.bdd.count(v_sel, nvars=n_vars - 1)
        # Number of configurations with the feature unselected
        v_unsel = bdd_model.bdd.let({feature: False}, u_func)
        nof_configs_var_unselected = bdd_model.bdd.count(v_unsel, nvars=n_vars - 1)

        # Randomly select or not the feature
        selected = random.choices([True, False],
                                  [nof_configs_var_selected, nof_configs_var_unselected],
                                  k=1)[0]

        # Update configuration and BDD node for the new feature
        values[feature] = selected
        u_func = bdd_model.bdd.let({feature: selected}, u_func)

        n_vars -= 1
    return Configuration(values)



from typing import Optional

from dd.autoref import BDD, Function

from famapy.core.models import VariabilityModel

from famapy.metamodels.bdd_metamodel.models.utils.txtcnf import TextCNFNotation, CNFLogicConnective


class BDDModel(VariabilityModel):
    """A Binary Decision Diagram (BDD) representation of the feature model.

    It relies on the dd library: https://pypi.org/project/dd/
    """

    CNF_NOTATION = TextCNFNotation.JAVA_SHORT
    NOT = CNF_NOTATION.value[CNFLogicConnective.NOT]
    AND = CNF_NOTATION.value[CNFLogicConnective.AND]
    OR = CNF_NOTATION.value[CNFLogicConnective.OR]

    @staticmethod
    def get_extension() -> str:
        return 'bdd'

    def __init__(self) -> None:
        self.bdd = BDD()  # BDD manager
        self.cnf_formula: Optional[str] = None 
        self.root = None
        self.variables: list[str] = []

    def from_textual_cnf(self, textual_cnf_formula: str, variables: list[str]) -> None:
        """Build the BDD from a textual representation of the CNF formula,
        and the list of variables."""
        self.cnf_formula = textual_cnf_formula
        self.variables = variables

        # Declare variables
        for var in self.variables:
            self.bdd.declare(var)

        # Build the BDD
        self.root = self.bdd.add_expr(self.cnf_formula)

        # Reorder variables
        # variable_order = self.bdd.vars 
        # var = self.bdd.var_at_level(0)
        # level = self.root.level
        # variable_order[self.root.var] = 0
        # variable_order[var] = level
        # self.bdd.reorder(variable_order)
        # self.root = self.bdd.var(self.bdd.var_at_level(0))

    def nof_nodes(self) -> int:
        """Return number of nodes in the BDD."""
        return len(self.bdd)

    @staticmethod
    def level(node: Function) -> int:
        """Return the level of the node. 

        Non-terminal nodes start at 0. 
        Terminal nodes have level `s' being the `s' the number of variables.
        """
        return node.level

    @staticmethod
    def index(node: Function) -> int:
        """Position (index) of the variable that labels the node `n` in the ordering.

        Indexes start at 1. 
        Terminal nodes (n0 and n1) have indexes `s + 1`, being `s' the number of variables.
        Note that index(n) = level(n) + 1.

        Example: node `n4` is labeled `B`, and `B` is in the 2nd position in ordering `[A,B,C]`,
        thus level(n4) = 2.
        """
        return node.level + 1

    @staticmethod
    def is_terminal_node(node: Function) -> bool:
        """Check if the node is a terminal node."""
        return node.var is None    

    @staticmethod
    def is_terminal_n1(node: Function) -> bool:
        """Check if the node is the terminal node 1 (n1)."""
        return node.var is None and node.node == 1

    @staticmethod
    def is_terminal_n0(node: Function) -> bool:
        """Check if the node is the terminal node 0 (n0)."""
        return node.var is None and node.node == -1

    @staticmethod
    def get_high_node(node: Function) -> Function:
        """Return the high (right, solid) node."""
        return node.high

    @staticmethod
    def get_low_node(node: Function) -> Function:
        """Return the low (left, dashed) node.

        If the arc is complemented it returns the negation of the left node.
        """
        return ~node.low if node.negated and node.low.var is not None else node.low



import itertools

from famapy.core.models import VariabilityModel
from famapy.core.transformations import ModelToModel
from famapy.metamodels.fm_metamodel.models.feature_model import (
    Constraint,
    Feature,
    Relation,
)
from famapy.metamodels.bdd_metamodel.models.bdd_model import BDDModel


class FmToBDD(ModelToModel):
    @staticmethod
    def get_source_extension() -> str:
        return 'fm'

    @staticmethod
    def get_destination_extension() -> str:
        return 'bdd'

    def __init__(self, source_model: VariabilityModel) -> None:
        self.source_model = source_model
        self.counter = 1
        self.destination_model = BDDModel()
        self.variables: dict[str, int] = {}
        self.features: dict[int, str] = {}
        self.clauses: list[list[int]] = []

    def add_feature(self, feature: Feature) -> None:
        if feature.name not in self.variables:
            self.variables[feature.name] = self.counter
            self.features[self.counter] = feature.name
            self.counter += 1

    def add_root(self, feature: Feature) -> None:
        var = self.variables.get(feature.name)
        if var is not None:
            self.clauses.append([var])

    def add_relation(self, relation: Relation) -> None:  # noqa: MC0001
        var_parent = self.variables.get(relation.parent.name)
        # TODO: fix too many nested blocks
        if var_parent is not None:  # pylint: disable=too-many-nested-blocks

            if relation.is_mandatory():
                var_child = self.variables.get(relation.children[0].name)
                if var_child is not None:
                    self.clauses.append([-1 * var_parent, var_child])
                    self.clauses.append([-1 * var_child, var_parent])

            elif relation.is_optional():
                var_child = self.variables.get(relation.children[0].name)
                if var_child is not None:
                    self.clauses.append([-1 * var_child, var_parent])

            elif relation.is_or():  # this is a 1 to n relatinship with multiple childs
                # add the first cnf child1 or child2 or ... or childN or no parent)

                # first elem of the constraint
                alt_cnf = [-1 * var_parent]
                for child in relation.children:
                    var_child = self.variables.get(child.name)
                    if var_child is not None:
                        alt_cnf.append(var_child)
                self.clauses.append(alt_cnf)

                for child in relation.children:
                    var_child = self.variables.get(child.name)
                    if var_child is not None:
                        self.clauses.append([-1 * var_child, var_parent])

            elif relation.is_alternative():
                # this is a 1 to 1 relatinship with multiple childs
                # add the first cnf child1 or child2 or ... or childN or no parent)

                # first elem of the constraint
                alt_cnf = [-1 * var_parent]
                for child in relation.children:
                    var_child = self.variables.get(child.name)
                    if var_child is not None:
                        alt_cnf.append(var_child)
                self.clauses.append(alt_cnf)

                for i, _ in enumerate(relation.children):
                    var_child_i = self.variables.get(relation.children[i].name)
                    if var_child_i is not None:
                        for j in range(i + 1, len(relation.children)):
                            if i != j:
                                var_child_j = self.variables.get(relation.children[j].name)
                                if var_child_j is not None:
                                    self.clauses.append([-1 * var_child_i, -1 * var_child_j])
                        self.clauses.append([-1 * var_child_i, var_parent])

            else:
                # This is a _min to _max relationship
                _min = relation.card_min
                _max = relation.card_max

                for val in range(len(relation.children) + 1):
                    if val < _min or val > _max:
                        # These sets are the combinations that shouldn't be in the res
                        # Let ¬A, B, C be one of your 0-paths.
                        # The relative clause will be (A ∨ ¬B ∨ ¬C).
                        # This first for loop is to combine when the parent is and
                        # the childs led to a 0-pathself.
                        for res in itertools.combinations(relation.children, val):
                            cnf = [-1 * var_parent]
                            for feat in relation.children:
                                var_feat = self.variables.get(feat.name)
                                if var_feat is not None:
                                    if feat in res:
                                        cnf.append(-1 * var_feat)
                                    else:
                                        cnf.append(var_feat)
                            self.clauses.append(cnf)
                    else:
                        # This first for loop is to combine when the parent is not
                        # and the childs led to a 1-pathself which is actually
                        # 0-path considering the parent.
                        for res in itertools.combinations(relation.children, val):
                            cnf = [var_parent]
                            for feat in relation.children:
                                var_feat = self.variables.get(feat.name)
                                if var_feat is not None:
                                    if feat in res:
                                        cnf.append(-1 * var_feat)
                                    else:
                                        cnf.append(var_feat)
                            self.clauses.append(cnf)

    def add_constraint(self, ctc: Constraint) -> None:
        clauses = ctc.ast.get_clauses()
        for clause in clauses:
            cls = []
            for term in clause:
                var_term = None
                if term.startswith('-'):
                    var_term = self.variables.get(term[1:])
                    if var_term is not None:
                        var_term = -1 * var_term
                else:
                    var_term = self.variables.get(term)
                if var_term is not None:
                    cls.append(var_term)
            self.clauses.append(cls)

    def transform(self) -> VariabilityModel:
        for feature in self.source_model.get_features():
            self.add_feature(feature)

        self.add_root(self.source_model.root)

        for relation in self.source_model.get_relations():
            self.add_relation(relation)

        for constraint in self.source_model.get_constraints():
            self.add_constraint(constraint)

        # Transform clauses to textual CNF notation required by the BDD
        not_connective = BDDModel.NOT
        or_connective = ' ' + BDDModel.OR + ' '
        and_connective = ' ' + BDDModel.AND + ' '
        cnf_list = []
        for clause in self.clauses:
            cnf_list.append('(' + or_connective.join(list(map(lambda l: 
                            not_connective + self.features[abs(l)] if l < 0 else 
                            self.features[abs(l)], clause))) + ')')

        cnf_formula = and_connective.join(cnf_list)
        self.destination_model.from_textual_cnf(cnf_formula, list(self.variables.keys()))

        return self.destination_model



from abc import abstractmethod
from typing import Any

from famapy.core.operations import Operation


class FeatureInclusionProbability(Operation):
    """The Feature Inclusion Probability (FIP) operation determines the probability 
    for a feature to be included in a valid product. 

    It obtains, for each feature, the proportion of valid products that include it.

    This operation in combination with the Product Distribution (PD) operation
    provides a clear picture of the products' homogeneity (i.e., how much does one product 
    differ from the others).
    The product's homogeneity is one the three core metrics that characterize 
    the complexity of a SPL.
    The other two core metrics are the number of features the SPL manages, 
    and the number of valid product that can be derived.

    Ref.: [Heradio et al. 2019. Supporting the Statistical Analysis of Variability Models. SPLC. 
    (https://doi.org/10.1109/ICSE.2019.00091)]
    """

    @abstractmethod
    def __init__(self) -> None:
        pass

    @abstractmethod
    def feature_inclusion_probability(self) -> dict[Any, float]:
        pass



from typing import Optional

from famapy.metamodels.configuration_metamodel.models.configuration import Configuration

from famapy.core.operations import ProductsNumber

from famapy.metamodels.bdd_metamodel.models.bdd_model import BDDModel


class BDDProductsNumber(ProductsNumber):
    """It computes the number of solutions of the BDD model.

    It also supports counting the solutions from a given partial configuration.
    """

    def __init__(self, partial_configuration: Configuration = None) -> None:
        self.result = 0
        self.bdd_model = None
        self.feature_model = None
        self.partial_configuration = partial_configuration

    def execute(self, model: BDDModel) -> 'BDDProductsNumber':
        self.bdd_model = model
        self.result = products_number(self.bdd_model, self.partial_configuration)
        return self

    def get_result(self) -> int:
        return self.result

    def get_products_number(self) -> int:
        return products_number(self.bdd_model, self.partial_configuration)


def products_number(bdd_model: BDDModel,
                    partial_configuration: Optional[Configuration] = None) -> int:
    if partial_configuration is None:
        u_func = bdd_model.root
        n_vars = len(bdd_model.variables)
    else:
        values = dict(partial_configuration.elements.items())
        u_func = bdd_model.bdd.let(values, bdd_model.root)
        n_vars = len(bdd_model.variables) - len(values)

    return bdd_model.bdd.count(u_func, nvars=n_vars)



from typing import Optional

from famapy.metamodels.configuration_metamodel.models.configuration import Configuration

from famapy.metamodels.bdd_metamodel.models import BDDModel
from famapy.metamodels.bdd_metamodel.operations.interfaces import ProductDistribution
from famapy.metamodels.bdd_metamodel.operations import BDDProducts


class BDDProductDistributionBF(ProductDistribution):
    """The Product Distribution (PD) algorithm determines the number of solutions
    having a given number of variables.

    This is a brute-force implementation that enumerates all solutions for accounting them.

    Ref.: [Heradio et al. 2019. Supporting the Statistical Analysis of Variability Models. SPLC.
    (https://doi.org/10.1109/ICSE.2019.00091)]
    """

    def __init__(self, partial_configuration: Configuration = None) -> None:
        self.result: list[int] = []
        self.bdd_model = None
        self.partial_configuration = partial_configuration

    def execute(self, model: BDDModel) -> 'BDDProductDistributionBF':
        self.bdd_model = model
        self.result = product_distribution(self.bdd_model, self.partial_configuration)
        return self

    def get_result(self) -> list[int]:
        return self.result

    def product_distribution(self) -> list[int]:
        return product_distribution(self.bdd_model, self.partial_configuration)



def product_distribution(bdd_model: BDDModel,
                         p_config: Optional[Configuration] = None) -> list[int]:
    """It accounts for how many solutions have no variables, one variable,
    two variables, ..., all variables.

    It enumerates all solutions and filters them.
    """
    products = BDDProducts(p_config).execute(bdd_model).get_result()
    dist: list[int] = []
    for i in range(len(bdd_model.variables) + 1):
        dist.append(sum(len(p.elements) == i for p in products))
    return dist



from .bdd_products_number import BDDProductsNumber, products_number
from .bdd_products import BDDProducts, products
from .bdd_sampling import BDDSampling, sample, random_configuration
from .bdd_product_distribution_bf import BDDProductDistributionBF, product_distribution
from .bdd_feature_inclusion_probability_bf import (
    BDDFeatureInclusionProbabilityBF, 
    feature_inclusion_probability
)


__all__ = ['BDDProductsNumber', 'products_number', 
           'BDDProducts', 'products',
           'BDDSampling', 'sample', 'random_configuration',
           'BDDProductDistributionBF', 'product_distribution',
           'BDDFeatureInclusionProbabilityBF', 'feature_inclusion_probability']


from abc import abstractmethod

from famapy.core.operations import Operation


class ProductDistribution(Operation):
    """The Product Distribution (PD) operation determines the number of products 
    having a given number of features.

    It accounts for how many products have no features, one features, 
    two features, ..., all features.

    This operation in combination with the Feature Inclusion Probability (FIP) operation
    provides a clear picture of the products' homogeneity (i.e., how much does one product 
    differ from the others).
    The product's homogeneity is one the three core metrics that characterize 
    the complexity of a SPL.
    The other two core metrics are the number of features the SPL manages, 
    and the number of valid product that can be derived.

    Ref.: [Heradio et al. 2019. Supporting the Statistical Analysis of Variability Models. SPLC. 
    (https://doi.org/10.1109/ICSE.2019.00091)]
    """

    @abstractmethod
    def __init__(self) -> None:
        pass

    @abstractmethod
    def product_distribution(self) -> list[int]:
        pass
